@@ -6,7 +6,7 @@
 
 __author__ = """luphord"""
 __email__ = """luphord@protonmail.com"""
-__version__ = """0.5.0"""
+__version__ = """0.7.0"""
 
 
 from argparse import ArgumentParser, Namespace
@@ -15,6 +15,7 @@ from itertools import groupby
 from typing import Final, Optional, Union, Mapping, Tuple, Set, Iterable, Callable
 from numbers import Real
 import numpy as np
+import pandas as pd
 from dataclasses import dataclass
 
 
@@ -23,29 +24,48 @@ _null_ccy = "NNN"
 ArrayLike = Union[np.ndarray, float]
 
 
-@dataclass
-class SimpleCashflows:
-    cashflows: np.ndarray
-    currencies: np.ndarray
-    dates: np.ndarray
-    nsim: int
+class SimpleCashflows(pd.DataFrame):
+    """Simple data structure to represent simulated cashflows.
+    Designed for external usage where an accumulated, simplified
+    view of the cashflow structure is sufficient.
+    Essentially a mapping from date and currency (row index)
+    to cashflow value per path (column index).
+    For convenience, this class is derived from `pandas.DataFrame`
+    to simplify further processing (such as plotting).
+    Can be created from `SimulatedCashflows`, but the reverse
+    direction is not possible.
+    """
 
-    def __init__(
-        self, cashflows: np.ndarray, currencies: np.ndarray, dates: np.ndarray
-    ):
+    @property
+    def _constructor(self) -> Callable:
+        return SimpleCashflows
+
+    @classmethod
+    def from_arrays(
+        cls, cashflows: np.ndarray, currencies: np.ndarray, dates: np.ndarray
+    ) -> "SimpleCashflows":
         assert cashflows.dtype == np.float64
         assert cashflows.ndim == 2, f"Array must have ndim 2, got {cashflows.ndim}"
         assert currencies.dtype == (np.unicode_, _ccy_letters)
         assert currencies.shape == (cashflows.shape[1],)
         assert dates.dtype == "datetime64[D]"
         assert dates.shape == (cashflows.shape[1],)
-        self.cashflows = cashflows
-        self.currencies = currencies
-        self.dates = dates
-        self.nsim = cashflows.shape[0]
+        index = pd.MultiIndex.from_arrays([dates, currencies])
+        return cls(pd.DataFrame(cashflows.T, index=index))
 
 
 class SimulatedCashflows:
+    """Complex data structure to represent simulated cashflows.
+    Designed for external usage where a detailed view
+    of the cashflow structure is desired.
+    Contains full information about the cashflow structure
+    indexed by datetime64[D] dates. Does not contain the integer
+    index values that correspond to these dates within the model.
+    Can be turned into `SimpleCashflows` via `to_simple_cashflows`.
+    Can be created from `IndexedCashflows`, but the reverse
+    direction is not possible.
+    """
+
     dtype: Final = np.dtype([("date", "datetime64[D]"), ("value", np.float64)])
     cashflows: Final[np.ndarray]
     currencies: Final[np.ndarray]
@@ -82,24 +102,36 @@ class SimulatedCashflows:
 
         for key, entries in groupby(sorted(self._split_by_date(), key=grpkey), grpkey):
             dt, ccy = key
-            yield sum(cf["value"] for cf, _, _ in entries), ccy, dt
+            sum_cf = sum(cf["value"] for cf, _, _ in entries)
+            assert isinstance(
+                sum_cf, np.ndarray
+            )  # appease mypy that sum_cf is actually an array, not 0
+            yield sum_cf, ccy, dt
 
     def to_simple_cashflows(self) -> SimpleCashflows:
         grouped_cf = list(self._group_cashflows())
         assert grouped_cf, "Got no cashflows to convert"
         numcf = len(grouped_cf)
-        cashflows = np.ndarray((self.nsim, numcf), dtype=np.float64)
-        currencies = np.ndarray((numcf,), dtype=(np.unicode_, _ccy_letters))
-        dates = np.ndarray((numcf,), dtype="datetime64[D]")
+        cashflows: np.ndarray = np.ndarray((self.nsim, numcf), dtype=np.float64)
+        currencies: np.ndarray = np.ndarray((numcf,), dtype=(np.unicode_, _ccy_letters))
+        dates: np.ndarray = np.ndarray((numcf,), dtype="datetime64[D]")
         for i, (cf, ccy, dt) in enumerate(grouped_cf):
             cashflows[:, i] = cf
             currencies[i] = ccy
             dates[i] = dt
-        return SimpleCashflows(cashflows, currencies, dates)
+        return SimpleCashflows.from_arrays(cashflows, currencies, dates)
 
 
 class IndexedCashflows:
-    dtype: Final = np.dtype([("index", np.int), ("value", np.float64)])
+    """Complex data structure to represent simulated cashflows
+    bound to a specific model.
+    Designed for internal usage within composable contracts.
+    Contains full information about the cashflow structure
+    indexed by integers corresponding to dates within the associated model.
+    Can be turned into `SimulatedCashflows` via `apply_index`.
+    """
+
+    dtype: Final = np.dtype([("index", np.int64), ("value", np.float64)])
     cashflows: Final[np.ndarray]
     currencies: Final[np.ndarray]
     dategrid: Final[np.ndarray]
@@ -166,16 +198,26 @@ class IndexedCashflows:
 
 
 class DateIndex:
+    """Integer index for time points (dates) within a model.
+    Index value may be different for each path.
+    Mathematically, this could be interpreted as the simulation
+    of a *stopping time* stochastic process.
+    Index value -1 indicates "never" (as in "this contract is never acquired").
+    """
+
     index: Final[np.ndarray]
     nsim: Final[int]
 
     def __init__(self, index: np.ndarray):
-        assert index.dtype == np.int
+        assert index.dtype == np.int64
         assert index.ndim == 1
         self.index = index
         self.nsim = index.size
 
     def index_column(self, observable: np.ndarray) -> np.ndarray:
+        """Apply this DateIndes to observable.
+        Essentially the same as observable[:, self.index],
+        but with proper handling for negative index values."""
         assert observable.ndim == 2
         assert observable.shape[0] == self.nsim
         assert (self.index < observable.shape[1]).all()
@@ -185,6 +227,8 @@ class DateIndex:
         return obs
 
     def next_after(self, obs: np.ndarray) -> "DateIndex":
+        """Create a new DateIndex where obs is True
+        for the first time after (including) this DateIndex."""
         assert obs.dtype == np.bool_
         assert obs.ndim == 2
         assert obs.shape[0] == self.nsim
@@ -198,6 +242,22 @@ class DateIndex:
         assert (idx < ndates).all()
         return DateIndex(idx)
 
+    def after_mask(self, ndates: int) -> np.ndarray:
+        """Return a boolean array of size (nsim, ndates)
+        where values after (including) this index are True."""
+        mask = np.repeat(
+            np.reshape(np.arange(ndates), (1, ndates)),
+            self.nsim,
+            axis=0,
+        ) >= np.reshape(self.index, (self.nsim, 1))
+        mask[self.index < 0, :] = False
+        return mask
+
+    def before_mask(self, ndates: int) -> np.ndarray:
+        """Return a boolean array of size (nsim, ndates)
+        where values before (excluding) this index are True."""
+        return ~self.after_mask(ndates)
+
 
 class TermStructuresModel(ABC):
     @abstractmethod
@@ -229,7 +289,7 @@ class Model:
         self.ndates = dategrid.size
         assert self.ndates > 0
         self.dategrid = np.reshape(dategrid, (self.ndates, 1))
-        assert numeraire.dtype == np.float
+        assert numeraire.dtype == np.float64
         assert numeraire.ndim == 2
         self.numeraire = numeraire
         self.nsim = numeraire.shape[0]
@@ -239,7 +299,7 @@ class Model:
         self.simulated_rates = simulated_rates
         for fxkey, val in simulated_fx.items():
             assert (
-                val.dtype == np.float
+                val.dtype == np.float64
             ), f"FX spot '{fxkey}' is of dtype {val.dtype}, expecting float"
             assert (
                 val.shape == self.shape
@@ -247,7 +307,7 @@ class Model:
         self.simulated_fx = simulated_fx
         for key, val in simulated_stocks.items():
             assert (
-                val.dtype == np.float
+                val.dtype == np.float64
             ), f"Stock '{key}' is of dtype {val.dtype}, expecting float"
             assert (
                 val.shape == self.shape
@@ -260,7 +320,7 @@ class Model:
 
     @property
     def eval_date_index(self) -> DateIndex:
-        return DateIndex(np.zeros((self.nsim,), dtype=np.int))
+        return DateIndex(np.zeros((self.nsim,), dtype=np.int64))
 
     @property
     def currencies(self) -> Set[str]:
@@ -362,7 +422,7 @@ class Model:
                 "Expecting IndexedCashflows or Contract, "
                 f"got {type(cashflows_or_contract)}"
             )
-        return self.discount(cf).sum(axis=1).mean(axis=0)
+        return float(self.discount(cf).sum(axis=1).mean(axis=0))
 
 
 class ObservableFloat(ABC):
@@ -370,7 +430,7 @@ class ObservableFloat(ABC):
     essentially a real-valued stochastic process"""
 
     @abstractmethod
-    def simulate(self, model: Model) -> np.ndarray:
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
         pass
 
     def __add__(self, other: Union["ObservableFloat", float]) -> "ObservableFloat":
@@ -379,7 +439,7 @@ class ObservableFloat(ABC):
         elif isinstance(other, Real):
             return Sum(self, KonstFloat(other))
         else:
-            raise TypeError(f"Expecting real number, got {other} of type {type(other)}")
+            return NotImplemented
 
     def __radd__(self, other: float) -> "ObservableFloat":
         return Sum(KonstFloat(other), self)
@@ -393,7 +453,7 @@ class ObservableFloat(ABC):
         elif isinstance(other, Real):
             return Sum(self, -KonstFloat(other))
         else:
-            raise TypeError(f"Expecting real number, got {other} of type {type(other)}")
+            return NotImplemented
 
     def __rsub__(self, other: float) -> "ObservableFloat":
         return Sum(KonstFloat(other), -self)
@@ -404,7 +464,7 @@ class ObservableFloat(ABC):
         elif isinstance(other, Real):
             return Product(self, KonstFloat(other))
         else:
-            raise TypeError(f"Expecting real number, got {other} of type {type(other)}")
+            return NotImplemented
 
     def __rmul__(self, other: float) -> "ObservableFloat":
         return Product(KonstFloat(other), self)
@@ -415,7 +475,7 @@ class ObservableFloat(ABC):
         elif isinstance(other, Real):
             return Quotient(self, KonstFloat(other))
         else:
-            raise TypeError(f"Expecting real number, got {other} of type {type(other)}")
+            return NotImplemented
 
     def __rtruediv__(self, other: float) -> "ObservableFloat":
         return Quotient(KonstFloat(other), self)
@@ -426,7 +486,7 @@ class ObservableFloat(ABC):
         elif isinstance(other, Real):
             return Power(self, KonstFloat(other))
         else:
-            raise TypeError(f"Expecting real number, got {other} of type {type(other)}")
+            return NotImplemented
 
     def __rpow__(self, other: float) -> "ObservableFloat":
         return Power(KonstFloat(other), self)
@@ -437,7 +497,7 @@ class ObservableFloat(ABC):
         elif isinstance(other, Real):
             return GreaterOrEqualThan(self, KonstFloat(other))
         else:
-            raise TypeError(f"Expecting real number, got {other} of type {type(other)}")
+            return NotImplemented
 
     def __gt__(self, other: Union["ObservableFloat", float]) -> "ObservableBool":
         if isinstance(other, ObservableFloat):
@@ -445,7 +505,7 @@ class ObservableFloat(ABC):
         elif isinstance(other, Real):
             return GreaterThan(self, KonstFloat(other))
         else:
-            raise TypeError(f"Expecting real number, got {other} of type {type(other)}")
+            return NotImplemented
 
     def __le__(self, other: Union["ObservableFloat", float]) -> "ObservableBool":
         return ~(self > other)
@@ -461,8 +521,10 @@ class Sum(ObservableFloat):
     observable1: ObservableFloat
     observable2: ObservableFloat
 
-    def simulate(self, model: Model) -> np.ndarray:
-        return self.observable1.simulate(model) + self.observable2.simulate(model)
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
+        return self.observable1.simulate(
+            first_observation_idx, model
+        ) + self.observable2.simulate(first_observation_idx, model)
 
     def __str__(self) -> str:
         return f"({self.observable1}) + ({self.observable2})"
@@ -474,8 +536,8 @@ class Minus(ObservableFloat):
 
     observable: ObservableFloat
 
-    def simulate(self, model: Model) -> np.ndarray:
-        return -self.observable.simulate(model)
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
+        return -self.observable.simulate(first_observation_idx, model)
 
     def __str__(self) -> str:
         return f"-{self.observable}"
@@ -488,8 +550,10 @@ class Product(ObservableFloat):
     observable1: ObservableFloat
     observable2: ObservableFloat
 
-    def simulate(self, model: Model) -> np.ndarray:
-        return self.observable1.simulate(model) * self.observable2.simulate(model)
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
+        return self.observable1.simulate(
+            first_observation_idx, model
+        ) * self.observable2.simulate(first_observation_idx, model)
 
     def __str__(self) -> str:
         return f"({self.observable1}) * ({self.observable2})"
@@ -502,8 +566,10 @@ class Quotient(ObservableFloat):
     observable1: ObservableFloat
     observable2: ObservableFloat
 
-    def simulate(self, model: Model) -> np.ndarray:
-        return self.observable1.simulate(model) / self.observable2.simulate(model)
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
+        return self.observable1.simulate(
+            first_observation_idx, model
+        ) / self.observable2.simulate(first_observation_idx, model)
 
     def __str__(self) -> str:
         return f"({self.observable1}) / ({self.observable2})"
@@ -516,26 +582,106 @@ class Power(ObservableFloat):
     observable1: ObservableFloat
     observable2: ObservableFloat
 
-    def simulate(self, model: Model) -> np.ndarray:
-        return self.observable1.simulate(model) ** self.observable2.simulate(model)
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
+        return self.observable1.simulate(
+            first_observation_idx, model
+        ) ** self.observable2.simulate(first_observation_idx, model)
 
     def __str__(self) -> str:
         return f"({self.observable1}) ** ({self.observable2})"
 
 
+@dataclass
+class Maximum(ObservableFloat):
+    """Equal to the maximum of two observables"""
+
+    observable1: ObservableFloat
+    observable2: ObservableFloat
+
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
+        return np.maximum(
+            self.observable1.simulate(first_observation_idx, model),
+            self.observable2.simulate(first_observation_idx, model),
+        )
+
+    def __str__(self) -> str:
+        return f"Maximum({self.observable1}, {self.observable2})"
+
+
+@dataclass
+class Minimum(ObservableFloat):
+    """Equal to the minimum of two observables"""
+
+    observable1: ObservableFloat
+    observable2: ObservableFloat
+
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
+        return np.minimum(
+            self.observable1.simulate(first_observation_idx, model),
+            self.observable2.simulate(first_observation_idx, model),
+        )
+
+    def __str__(self) -> str:
+        return f"Minimum({self.observable1}, {self.observable2})"
+
+
+def _simulate_accumulate(
+    accfn: np.ufunc,
+    observable: ObservableFloat,
+    first_observation_idx: DateIndex,
+    model: Model,
+) -> np.ndarray:
+    underlying = observable.simulate(first_observation_idx, model).copy()
+    running_accumulate = underlying.copy()
+    mask = first_observation_idx.before_mask(model.ndates)
+    running_accumulate[mask] = np.nan
+    running_accumulate = accfn.accumulate(running_accumulate, axis=1)
+    running_accumulate[mask] = underlying[mask]
+    return running_accumulate
+
+
+@dataclass
+class RunningMax(ObservableFloat):
+    """Running maximum of observable over time, seen from first_observation_idx."""
+
+    observable: ObservableFloat
+
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
+        return _simulate_accumulate(
+            np.fmax, self.observable, first_observation_idx, model
+        )
+
+    def __str__(self) -> str:
+        return f"RunningMax({self.observable})"
+
+
+@dataclass
+class RunningMin(ObservableFloat):
+    """Running minimum of observable over time, seen from first_observation_idx."""
+
+    observable: ObservableFloat
+
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
+        return _simulate_accumulate(
+            np.fmin, self.observable, first_observation_idx, model
+        )
+
+    def __str__(self) -> str:
+        return f"RunningMin({self.observable})"
+
+
 @dataclass
 class FixedAfter(ObservableFloat):
     """Equal to observable, but remains constant as soon as
-    fixing_condition becomes true (seen from the beginning,
-    not from acquisition date)"""
+    fixing_condition becomes true after (including) first_observation_idx."""
 
     fixing_condition: "ObservableBool"
     observable: ObservableFloat
 
-    def simulate(self, model: Model) -> np.ndarray:
-        underlying = self.observable.simulate(model).copy()
-        fixing_times = self.fixing_condition.simulate(model)
-        fixing_idx = model.eval_date_index.next_after(fixing_times)
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
+        underlying = self.observable.simulate(first_observation_idx, model).copy()
+        fixing_times = self.fixing_condition.simulate(first_observation_idx, model)
+        fixing_idx = first_observation_idx.next_after(fixing_times)
         for i in range(1, underlying.shape[1]):
             fixing_mask = (i > fixing_idx.index) & (fixing_idx.index >= 0)
             underlying[fixing_mask, i] = underlying[fixing_mask, i - 1]
@@ -551,7 +697,7 @@ class Stock(ObservableFloat):
 
     identifier: str
 
-    def simulate(self, model: Model) -> np.ndarray:
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
         return model.simulated_stocks[self.identifier]
 
     def __str__(self) -> str:
@@ -566,7 +712,7 @@ class FX(ObservableFloat):
     base_currency: str
     counter_currency: str
 
-    def simulate(self, model: Model) -> np.ndarray:
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
         return model.get_simulated_fx(self.base_currency, self.counter_currency)
 
     def __str__(self) -> str:
@@ -581,7 +727,7 @@ class LinearRate(ObservableFloat):
     currency: str
     frequency: str
 
-    def simulate(self, model: Model) -> np.ndarray:
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
         return model.simulated_rates[self.currency].linear_rate(self.frequency)
 
     def __str__(self) -> str:
@@ -594,8 +740,8 @@ class KonstFloat(ObservableFloat):
 
     constant: float
 
-    def simulate(self, model: Model) -> np.ndarray:
-        return self.constant * np.ones(model.shape, dtype=np.float)
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
+        return self.constant * np.ones(model.shape, dtype=np.float64)
 
     def __str__(self) -> str:
         return str(self.constant)
@@ -605,7 +751,7 @@ class ObservableBool(ABC):
     """Abstract base class for all observables of underlying type bool"""
 
     @abstractmethod
-    def simulate(self, model: Model) -> np.ndarray:
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
         pass
 
     def __invert__(self) -> "ObservableBool":
@@ -624,8 +770,8 @@ class Not(ObservableBool):
 
     observable: ObservableBool
 
-    def simulate(self, model: Model) -> np.ndarray:
-        return ~self.observable.simulate(model)
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
+        return ~self.observable.simulate(first_observation_idx, model)
 
     def __str__(self) -> str:
         return f"~({self.observable})"
@@ -638,8 +784,10 @@ class AndObservable(ObservableBool):
     observable1: ObservableBool
     observable2: ObservableBool
 
-    def simulate(self, model: Model) -> np.ndarray:
-        return self.observable1.simulate(model) & self.observable2.simulate(model)
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
+        return self.observable1.simulate(
+            first_observation_idx, model
+        ) & self.observable2.simulate(first_observation_idx, model)
 
     def __str__(self) -> str:
         return f"({self.observable1}) & ({self.observable2})"
@@ -652,8 +800,10 @@ class OrObservable(ObservableBool):
     observable1: ObservableBool
     observable2: ObservableBool
 
-    def simulate(self, model: Model) -> np.ndarray:
-        return self.observable1.simulate(model) | self.observable2.simulate(model)
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
+        return self.observable1.simulate(
+            first_observation_idx, model
+        ) | self.observable2.simulate(first_observation_idx, model)
 
     def __str__(self) -> str:
         return f"({self.observable1}) | ({self.observable2})"
@@ -666,8 +816,10 @@ class GreaterOrEqualThan(ObservableBool):
     observable1: ObservableFloat
     observable2: ObservableFloat
 
-    def simulate(self, model: Model) -> np.ndarray:
-        return self.observable1.simulate(model) >= self.observable2.simulate(model)
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
+        return self.observable1.simulate(
+            first_observation_idx, model
+        ) >= self.observable2.simulate(first_observation_idx, model)
 
     def __str__(self) -> str:
         return f"{self.observable1} >= {self.observable2}"
@@ -680,8 +832,10 @@ class GreaterThan(ObservableBool):
     observable1: ObservableFloat
     observable2: ObservableFloat
 
-    def simulate(self, model: Model) -> np.ndarray:
-        return self.observable1.simulate(model) > self.observable2.simulate(model)
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
+        return self.observable1.simulate(
+            first_observation_idx, model
+        ) > self.observable2.simulate(first_observation_idx, model)
 
     def __str__(self) -> str:
         return f"{self.observable1} > {self.observable2}"
@@ -693,9 +847,10 @@ class At(ObservableBool):
 
     date: np.datetime64
 
-    def simulate(self, model: Model) -> np.ndarray:
+    def simulate(self, first_observation_idx: DateIndex, model: Model) -> np.ndarray:
         mask = (model.dategrid == self.date).reshape((1, model.ndates))
         assert mask.any(), f"{self.date} not contained in dategrid"
+        # ToDo: Should we assert that self.date is after first_observation_idx?
         return np.repeat(mask, model.nsim, axis=0)
 
     def __str__(self) -> str:
@@ -711,6 +866,29 @@ class Contract(ABC):
     ) -> IndexedCashflows:
         pass
 
+    def __add__(self, other: "Contract") -> "Contract":
+        return And(self, other)
+
+    def __neg__(self) -> "Contract":
+        return Give(self)
+
+    def __sub__(self, other: "Contract") -> "Contract":
+        return And(self, Give(other))
+
+    def __mul__(self, observable: Union[ObservableFloat, float]) -> "Contract":
+        if isinstance(observable, ObservableFloat):
+            return Scale(observable, self)
+        elif isinstance(observable, Real):
+            return Scale(KonstFloat(observable), self)
+        else:
+            return NotImplemented
+
+    def __rmul__(self, observable: Union[ObservableFloat, float]) -> "Contract":
+        return self.__mul__(observable)
+
+    def __or__(self, other: "Contract") -> "Contract":
+        return Or(self, other)
+
 
 @dataclass
 class Zero(Contract):
@@ -830,7 +1008,9 @@ class Cond(Contract):
     def generate_cashflows(
         self, acquisition_idx: DateIndex, model: Model
     ) -> IndexedCashflows:
-        obs = acquisition_idx.index_column(self.observable.simulate(model))
+        obs = acquisition_idx.index_column(
+            self.observable.simulate(acquisition_idx, model)
+        )
         never = acquisition_idx.index < 0
         cf1 = self.contract1.generate_cashflows(acquisition_idx, model)
         cf1.cashflows["value"][~obs, :] = 0
@@ -858,7 +1038,9 @@ class Scale(Contract):
         self, acquisition_idx: DateIndex, model: Model
     ) -> IndexedCashflows:
         cf = self.contract.generate_cashflows(acquisition_idx, model)
-        obs = acquisition_idx.index_column(self.observable.simulate(model))
+        obs = acquisition_idx.index_column(
+            self.observable.simulate(acquisition_idx, model)
+        )
         assert cf.cashflows.ndim == 2
         assert cf.cashflows.shape[0] == model.nsim
         assert obs.ndim == 1
@@ -880,7 +1062,9 @@ class When(Contract):
     def generate_cashflows(
         self, acquisition_idx: DateIndex, model: Model
     ) -> IndexedCashflows:
-        idx = acquisition_idx.next_after(self.observable.simulate(model))
+        idx = acquisition_idx.next_after(
+            self.observable.simulate(acquisition_idx, model)
+        )
         return self.contract.generate_cashflows(idx, model)
 
     def __str__(self) -> str:
@@ -916,7 +1100,9 @@ class Until(Contract):
         self, acquisition_idx: DateIndex, model: Model
     ) -> IndexedCashflows:
         cf = self.contract.generate_cashflows(acquisition_idx, model)
-        ko_idx = acquisition_idx.next_after(self.observable.simulate(model))
+        ko_idx = acquisition_idx.next_after(
+            self.observable.simulate(acquisition_idx, model)
+        )
         return cf.zero_after(ko_idx)
 
     def __str__(self) -> str:
@@ -944,9 +1130,7 @@ class ZeroCouponBond(ResolvableContract):
     currency: str
 
     def resolve(self) -> Contract:
-        return When(
-            At(self.maturity), Scale(KonstFloat(self.notional), One(self.currency))
-        )
+        return When(At(self.maturity), self.notional * One(self.currency))
 
 
 @dataclass
@@ -955,14 +1139,14 @@ class EuropeanOption(ResolvableContract):
     contract: Contract
 
     def resolve(self) -> Contract:
-        return When(At(self.maturity), Or(self.contract, Zero()))
+        return When(At(self.maturity), self.contract | Zero())
 
 
 class StochasticProcess(ABC):
     """Base class for stochastic processes."""
 
     @abstractmethod
-    def simulate(self, t: np.array, n: int, rnd: np.random.RandomState) -> np.array:
+    def simulate(self, t: np.ndarray, n: int, rnd: np.random.RandomState) -> np.ndarray:
         pass
 
     @abstractmethod
@@ -983,6 +1167,7 @@ class StochasticProcess(ABC):
         target_stddev = self.stddev(t)
         actual_mean = paths.mean(axis=0)
         actual_stddev = paths.std(axis=0)
+        assert isinstance(actual_stddev, np.ndarray)
         assert (
             target_stddev[actual_stddev == 0] == 0
         ).all(), "Cannot scale actual stddev of zero to any other value than zero"
@@ -990,8 +1175,8 @@ class StochasticProcess(ABC):
         return (paths - actual_mean) / actual_stddev * target_stddev + target_mean
 
     def simulate_with_moment_matching(
-        self, t: np.array, n: int, rnd: np.random.RandomState
-    ) -> np.array:
+        self, t: np.ndarray, n: int, rnd: np.random.RandomState
+    ) -> np.ndarray:
         return self.moment_match(t, self.simulate(t, n, rnd))
 
 
@@ -999,12 +1184,14 @@ class BrownianMotion(StochasticProcess):
     """Brownian Motion (Wiener Process) with optional drift."""
 
     def __init__(
-        self, mu_t: Callable[[float], float] = lambda t: 0, sigma: float = 1.0
+        self,
+        mu_t: Callable[[np.ndarray], np.ndarray] = np.zeros_like,
+        sigma: float = 1.0,
     ):
         self.mu_t = mu_t
         self.sigma = sigma
 
-    def simulate(self, t: np.array, n: int, rnd: np.random.RandomState) -> np.array:
+    def simulate(self, t: np.ndarray, n: int, rnd: np.random.RandomState) -> np.ndarray:
         assert t.ndim == 1, "One dimensional time vector required"
         assert t.size > 0, "At least one time point is required"
         dt = np.concatenate((t[0:1], np.diff(t)))
@@ -1018,19 +1205,21 @@ class BrownianMotion(StochasticProcess):
         return self.mu_t(t)
 
     def stddev(self, t: np.ndarray) -> np.ndarray:
-        return np.sqrt(self.sigma ** 2 * t)
+        return np.sqrt(self.sigma**2 * t)
 
 
 class GeometricBrownianMotion(StochasticProcess):
     """Geometric Brownian Motion.(with optional drift)."""
 
     def __init__(
-        self, mu_t: Callable[[float], float] = lambda t: 0, sigma: float = 1.0
+        self,
+        mu_t: Callable[[np.ndarray], np.ndarray] = np.zeros_like,
+        sigma: float = 1.0,
     ):
         self.mu_t = mu_t
         self.sigma = sigma
 
-    def simulate(self, t: np.array, n: int, rnd: np.random.RandomState) -> np.array:
+    def simulate(self, t: np.ndarray, n: int, rnd: np.random.RandomState) -> np.ndarray:
         assert t.ndim == 1, "One dimensional time vector required"
         assert t.size > 0, "At least one time point is required"
         dt = np.concatenate((t[0:1], np.diff(t)))
@@ -1038,13 +1227,13 @@ class GeometricBrownianMotion(StochasticProcess):
         # transposed simulation for automatic broadcasting
         dW = (rnd.normal(size=(t.size, n)).T * np.sqrt(dt)).T
         W = np.cumsum(dW, axis=0)
-        return np.exp(self.sigma * W.T + self.mu_t(t) - self.sigma ** 2 / 2 * t)
+        return np.exp(self.sigma * W.T + self.mu_t(t) - self.sigma**2 / 2 * t)
 
     def expected(self, t: np.ndarray) -> np.ndarray:
         return np.exp(self.mu_t(t))
 
     def stddev(self, t: np.ndarray) -> np.ndarray:
-        return np.sqrt(np.exp(2 * self.mu_t(t)) * (np.exp(self.sigma ** 2 * t) - 1))
+        return np.sqrt(np.exp(2 * self.mu_t(t)) * (np.exp(self.sigma**2 * t) - 1))
 
 
 def _get_year_fractions(dategrid: np.ndarray) -> np.ndarray:
@@ -1056,7 +1245,7 @@ def simulate_equity_black_scholes_model(
     stock: str,
     currency: str,
     S0: float,
-    dategrid: np.array,
+    dategrid: np.ndarray,
     sigma: float,
     r: float,
     n: int,
@@ -1080,19 +1269,19 @@ class HoLeeModel(TermStructuresModel):
 
     h: Final[float] = np.sqrt(np.finfo(float).eps)
 
-    dategrid: Final[np.array]
-    yearfractions: Final[np.array]
+    dategrid: Final[np.ndarray]
+    yearfractions: Final[np.ndarray]
     shortrates: Final[np.ndarray]
     discount_curve: Optional[
-        Callable[[float], float]
+        Callable[[np.ndarray], np.ndarray]
     ]  # mypy requires hack, see https://ogy.de/mypy-callable-members
     sigma: Final[float]
     _mu_t: Final[np.ndarray]
 
     def __init__(
         self,
-        dategrid: np.array,
-        discount_curve: Callable[[float], float],
+        dategrid: np.ndarray,
+        discount_curve: Callable[[np.ndarray], np.ndarray],
         sigma: float,
         n: int,
         rnd: np.random.RandomState,
@@ -1113,16 +1302,18 @@ class HoLeeModel(TermStructuresModel):
             np.log(self.discount_curve(self.yearfractions + self.h))
             - np.log(self.discount_curve(self.yearfractions))
         ) / self.h
-        return self.sigma ** 2 * self.yearfractions ** 2 / 2 - dlogBond
+        return self.sigma**2 * self.yearfractions**2 / 2 - dlogBond
 
-    def _integral_mu_t(self, T: float) -> np.ndarray:
+    def _integral_mu_t(self, T: np.ndarray) -> np.ndarray:
         """Integral over mu_t (i.e. double integral over theta)
         from model grid to T"""
         assert T >= self.yearfractions[0]
         assert T <= self.yearfractions[-1]
         t = np.linspace(self.yearfractions[0], self.yearfractions[-1], num=100)
         int_mu = np.cumsum(self.mu_t(t) * np.diff(t, prepend=self.yearfractions[0] - 1))
-        int_mu_fn: Callable[[float], float] = lambda tau: np.interp(tau, t, int_mu)
+        int_mu_fn: Callable[[np.ndarray], np.ndarray] = lambda tau: np.interp(
+            tau, t, int_mu
+        )
         return int_mu_fn(T) - int_mu_fn(self.yearfractions)
 
     def _simulate(
@@ -1148,14 +1339,14 @@ class HoLeeModel(TermStructuresModel):
         r = np.hstack((r_before_0, self.shortrates[:, :-1]))
         return np.exp(-np.cumsum(r * dt, axis=1))
 
-    def bond_prices(self, T: float) -> np.ndarray:
+    def bond_prices(self, T: np.ndarray) -> np.ndarray:
         """Stochastic T-bond prices implied by the model,
         quoted for all times even after T"""
         t = self.yearfractions
         lb = (
             -(T - t) * self.shortrates
             - self._integral_mu_t(T)
-            + self.sigma ** 2 / 6 * (T - t) ** 3
+            + self.sigma**2 / 6 * (T - t) ** 3
         )
         return np.exp(lb)
 
